/*
 * $Header: /home/cvspublic/jakarta-slide/src/share/org/apache/slide/content/ContentImpl.java,v 1.27 2001/05/16 12:04:24 juergen Exp $
 * $Revision: 1.27 $
 * $Date: 2001/05/16 12:04:24 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */

package org.apache.slide.content;

import java.util.Enumeration;
import java.util.Date;
import java.util.Vector;
import java.util.Hashtable;
import org.apache.slide.common.*;
import org.apache.slide.structure.*;
import org.apache.slide.lock.*;
import org.apache.slide.security.*;
import org.apache.slide.authenticate.CredentialsToken;

/**
 * Implementation of the content interface.
 *
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 * @version $Revision: 1.27 $
 */
public final class ContentImpl implements Content {
    
    
    // -------------------------------------------------------------- Constants
    
    
    protected static final int PRE_STORE = 0;
    protected static final int POST_STORE = 1;
    protected static final int POST_RETRIEVE = 2;
    
    
    // ----------------------------------------------------------- Constructors
    
    
    /**
     * Constructor.
     *
     * @param namespace Namespace
     * @param namespaceConfig Namespace configuration
     * @param securityHelper Security helper
     * @param dataHelper Data helper
     * @param lockHelper lockHelper
     */
    public ContentImpl(Namespace namespace, NamespaceConfig namespaceConfig,
                       Security securityHelper, Structure structureHelper,
                       Lock lockHelper) {
        this.namespace = namespace;
        this.namespaceConfig = namespaceConfig;
        this.securityHelper = securityHelper;
        this.structureHelper = structureHelper;
        this.lockHelper = lockHelper;
    }
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    /**
     * Namespace.
     */
    private Namespace namespace;
    
    
    /**
     * Namespace configuration.
     */
    private NamespaceConfig namespaceConfig;
    
    
    /**
     * Security helper.
     */
    private Security securityHelper;
    
    
    /**
     * Structure helper.
     */
    private Structure structureHelper;
    
    
    /**
     * Lock helper.
     */
    private Lock lockHelper;
    
    
    // -------------------------------------------------------- Content Methods
    
    
    /**
     * Retrieve revision descriptors.
     *
     * @param strUri Uri
     * @return NodeRevisionDescriptors
     */
    public NodeRevisionDescriptors retrieve(SlideToken token,
                                            String strUri)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        ObjectLockedException {
        ObjectNode associatedObject = structureHelper.retrieve(token, strUri);
        
        // Checking security and locking
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getReadRevisionMetadataAction());
        lockHelper.checkLock(token, associatedObject,
                             namespaceConfig.getReadRevisionMetadataAction());
        
        Uri objectUri = namespace.getUri(token, strUri);
        NodeRevisionDescriptors revisionDescriptors = null;
        try {
            revisionDescriptors = objectUri.getStore()
                .retrieveRevisionDescriptors(objectUri);
        } catch (RevisionDescriptorNotFoundException e) {
            // No revision descriptors. We have to create some.
            revisionDescriptors = new NodeRevisionDescriptors();
            revisionDescriptors.setUri(objectUri.toString());
            objectUri.getStore()
                .createRevisionDescriptors(objectUri, revisionDescriptors);
        }
        return revisionDescriptors;
    }
    
    
    /**
     * Retrieve revision descriptor.
     *
     * @param revisionDescriptors Node revision descriptors
     * @param revisionNumber Node revision number
     */
    public NodeRevisionDescriptor retrieve
        (SlideToken token, NodeRevisionDescriptors revisionDescriptors,
         String branch)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionDescriptorNotFoundException, ObjectLockedException,
        BranchNotFoundException, NodeNotVersionedException {
        
        Uri objectUri = namespace.getUri(token, revisionDescriptors.getUri());
        
        NodeRevisionDescriptors realRevisionDescriptors = objectUri.getStore()
            .retrieveRevisionDescriptors(objectUri);
        
        if (!realRevisionDescriptors.isVersioned()) {
            // Invalid function call : we try to create a revision, but the
            // descriptors won't allow it
            throw new NodeNotVersionedException
                (realRevisionDescriptors.getUri().toString());
        }
        
        // Retrieving latest revision numbers
        NodeRevisionNumber branchLatestRevisionNumber =
            realRevisionDescriptors.getLatestRevision(branch);
        
        if (branchLatestRevisionNumber == null) {
            throw new BranchNotFoundException
                (realRevisionDescriptors.getUri().toString(), branch);
        }
        
        return retrieve(token, realRevisionDescriptors,
                        branchLatestRevisionNumber);
        
    }
    
    
    /**
     * Retrieve revision descriptor.
     *
     * @param revisionDescriptors Node revision descriptors
     * @param revisionNumber Node revision number
     */
    public NodeRevisionDescriptor retrieve
        (SlideToken token, NodeRevisionDescriptors revisionDescriptors,
         NodeRevisionNumber revisionNumber)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionDescriptorNotFoundException, ObjectLockedException {
        
        ObjectNode associatedObject = structureHelper
            .retrieve(token, revisionDescriptors.getUri());
        
        // Checking security and locking
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getReadRevisionMetadataAction());
        lockHelper.checkLock(token, associatedObject,
                             namespaceConfig.getReadRevisionMetadataAction());
        
        Uri objectUri = namespace.getUri(token, revisionDescriptors.getUri());
        
        NodeRevisionDescriptor revisionDescriptor =
            objectUri.getStore().retrieveRevisionDescriptor
            (objectUri, revisionNumber);
        
        // Invoke interceptors
        invokeInterceptors(token, revisionDescriptors, revisionDescriptor,
                           null, POST_RETRIEVE);
        
        return revisionDescriptor;
        
    }
    
    
    /**
     * Retrieve revision descriptor from the latest revision
     * in the main branch.
     *
     * @param revisionDescriptors Node revision descriptors
     */
    public NodeRevisionDescriptor retrieve
        (SlideToken token, NodeRevisionDescriptors revisionDescriptors)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionDescriptorNotFoundException, ObjectLockedException {
        
        Uri objectUri = namespace.getUri(token, revisionDescriptors.getUri());
        
        NodeRevisionDescriptors realRevisionDescriptors = objectUri.getStore()
            .retrieveRevisionDescriptors(objectUri);
        
        return retrieve(token, revisionDescriptors,
                        realRevisionDescriptors.getLatestRevision());
        
    }
    
    
    /**
     * Retrieve revision content.
     *
     * @param revisionDescriptors Node revision descriptors
     * @param revisionDescriptor Node revision descriptor
     */
    public NodeRevisionContent retrieve
        (SlideToken token, NodeRevisionDescriptors revisionDescriptors,
         NodeRevisionDescriptor revisionDescriptor)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionNotFoundException, RevisionContentNotFoundException,
        ObjectLockedException {
        return retrieve(token, revisionDescriptors.getUri(),
                        revisionDescriptor);
    }
    
    
    /**
     * Retrieve revision content.
     *
     * @param strUri Uri
     * @param revisionDescriptor Node revision descriptor
     */
    public NodeRevisionContent retrieve
        (SlideToken token, String strUri,
         NodeRevisionDescriptor revisionDescriptor)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionNotFoundException, RevisionContentNotFoundException,
        ObjectLockedException {
        
        ObjectNode associatedObject = structureHelper.retrieve(token, strUri);
        
        // Checking security and locking
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getReadRevisionContentAction());
        lockHelper.checkLock(token, associatedObject,
                             namespaceConfig.getReadRevisionContentAction());
        
        Uri objectUri = namespace.getUri(token, strUri);
        NodeRevisionContent revisionContent =
            objectUri.getStore().retrieveRevisionContent(objectUri,
                                                         revisionDescriptor);
        
        // Invoke interceptors
        invokeInterceptors(token, null, revisionDescriptor,
                           revisionContent, POST_RETRIEVE);
        
        return revisionContent;
    }
    
    
    /**
     * Create new revision descriptors.
     *
     * @param strUri Uri
     * @param isVersioned true is the resource is versioned
     */
    public void create(SlideToken token, String strUri,
                       boolean isVersioned)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        ObjectLockedException {
        
        // Retrieve the associated object
        ObjectNode associatedObject = structureHelper.retrieve(token, strUri);
        
        // Next we do a security check and a locking check for modifyRevisions
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getCreateRevisionMetadataAction());
        lockHelper.checkLock
            (token, associatedObject,
             namespaceConfig.getCreateRevisionMetadataAction());
        if (namespaceConfig.getCreateRevisionMetadataAction() !=
            namespaceConfig.getCreateRevisionContentAction()) {
            securityHelper.checkCredentials
                (token, associatedObject,
                 namespaceConfig.getCreateRevisionContentAction());
            lockHelper.checkLock
                (token, associatedObject,
                 namespaceConfig.getCreateRevisionContentAction());
        }
        
        Uri objectUri = namespace.getUri(token, strUri);
        
        NodeRevisionDescriptors revisionDescriptors =
            new NodeRevisionDescriptors(isVersioned);
        revisionDescriptors.setUri(strUri);
        objectUri.getStore()
            .createRevisionDescriptors(objectUri, revisionDescriptors);
        
    }
    
    
    /**
     * Create new revision in main branch.
     *
     * @param strUri Uri
     * @param revisionDescriptor New Node revision descriptor
     * @param revisionContent New Node revision content
     */
    public void create(SlideToken token, String strUri,
                       NodeRevisionDescriptor revisionDescriptor,
                       NodeRevisionContent revisionContent)
        throws ObjectNotFoundException, AccessDeniedException,
        RevisionAlreadyExistException, LinkedObjectNotFoundException,
        ServiceAccessException, ObjectLockedException {
        
        // Retrieve the associated object
        ObjectNode associatedObject = structureHelper.retrieve(token, strUri);
        
        // Next we do a security check and a locking check for modifyRevisions
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getCreateRevisionMetadataAction());
        lockHelper.checkLock
            (token, associatedObject,
             namespaceConfig.getCreateRevisionMetadataAction());
        if (namespaceConfig.getCreateRevisionMetadataAction() !=
            namespaceConfig.getCreateRevisionContentAction()) {
            securityHelper.checkCredentials
                (token, associatedObject,
                 namespaceConfig.getCreateRevisionContentAction());
            lockHelper.checkLock
                (token, associatedObject,
                 namespaceConfig.getCreateRevisionContentAction());
        }
        
        setDefaultProperties(associatedObject, revisionDescriptor);
        // set the creation date if not already set
        if (revisionDescriptor.getCreationDate() == null) {
            revisionDescriptor.setCreationDate(new Date());
        }
        // set the display name (in case of copy)
        String resourceName = strUri;
        int lastSlash = resourceName.lastIndexOf('/');
        if (lastSlash != -1)
            resourceName = resourceName.substring(lastSlash + 1);
        revisionDescriptor.setName(resourceName);
        
        Uri objectUri = namespace.getUri(token, strUri);
        
        NodeRevisionDescriptors revisionDescriptors = null;
        try {
            revisionDescriptors = objectUri.getStore()
                .retrieveRevisionDescriptors(objectUri);
        } catch (RevisionDescriptorNotFoundException e) {
            // No revision descriptors. We have to create some.
            revisionDescriptors = new NodeRevisionDescriptors();
            revisionDescriptors.setUri(strUri);
            objectUri.getStore()
                .createRevisionDescriptors(objectUri, revisionDescriptors);
        }
        
        // Retrieve the latest revision from the descriptor,
        // unless there is no revisions. We generate a new revision number,
        // basing on an existing revision, if any.
        NodeRevisionNumber newRevisionNumber = null;
        if (revisionDescriptors.isVersioned()) {
            
            if (revisionDescriptors.hasRevisions()) {
                newRevisionNumber = new NodeRevisionNumber
                    (revisionDescriptors.getLatestRevision());
                revisionDescriptors
                    .addSuccessor(revisionDescriptors.getLatestRevision(),
                                  newRevisionNumber);
                revisionDescriptors
                    .setSuccessors(newRevisionNumber, new Vector());
            } else {
                newRevisionNumber = new NodeRevisionNumber();
                revisionDescriptors
                    .setSuccessors(newRevisionNumber, new Vector());
            }
            // We now set the newly created revision as the latest revison
            revisionDescriptors.setLatestRevision(newRevisionNumber);
            
            // We update the descriptor
            revisionDescriptor.setRevisionNumber(newRevisionNumber);
            
            // Invoke interceptors
            invokeInterceptors(token, revisionDescriptors, revisionDescriptor,
                               revisionContent, PRE_STORE);
            
            if (revisionContent != null) {
                // Storing the new revision contents
                objectUri.getStore()
                    .createRevisionContent(objectUri, revisionDescriptor,
                                           revisionContent);
            }
            // Now creating the revision desriptor in the store
            objectUri.getStore()
                .createRevisionDescriptor(objectUri, revisionDescriptor);
            
        } else {
            // We don't use versioning for this object.
            // Two options :
            // - The object already has one (and only one) revision,
            //   so we update it
            // - The object dooesn't have any revisions right now, so we create
            //   the initial revision
            newRevisionNumber = new NodeRevisionNumber();
            revisionDescriptor.setRevisionNumber(newRevisionNumber);
            
            if (!revisionDescriptors.hasRevisions()) {
                
                // Invoke interceptors
                invokeInterceptors(token, revisionDescriptors,
                                   revisionDescriptor,
                                   revisionContent, PRE_STORE);
                
                if (revisionContent != null) {
                    // Storing the new revision contents
                    objectUri.getStore()
                        .createRevisionContent(objectUri, revisionDescriptor,
                                               revisionContent);
                }
                // Now creating the revision desriptor in the store
                objectUri.getStore()
                    .createRevisionDescriptor(objectUri, revisionDescriptor);
                
            } else {
                
                try {
                    { 
                        // merge the new received properties into the 
                        // revisionDescriptor

                        // We update the descriptor's properties
                        NodeRevisionDescriptor oldRevisionDescriptor =
                            objectUri.getStore()
                            .retrieveRevisionDescriptor
                            (objectUri, newRevisionNumber);
                        Enumeration newPropertiesList =
                            revisionDescriptor.enumerateProperties();
                        while (newPropertiesList.hasMoreElements()) {
                            oldRevisionDescriptor
                                .setProperty((NodeProperty) newPropertiesList
                                             .nextElement() );
                        }
                        
                        // now use the merged revision descriptor
                        revisionDescriptor = oldRevisionDescriptor;
                    } // end of merge
                    
                    // Invoke interceptors
                    invokeInterceptors(token, revisionDescriptors,
                                       revisionDescriptor,
                                       revisionContent, PRE_STORE);
                    
                    if (revisionContent != null) {
                        // Storing the new revision contents
                        try {
                            objectUri.getStore()
                                .storeRevisionContent(objectUri,
                                                      revisionDescriptor,
                                                      revisionContent);
                        } catch (RevisionNotFoundException e) {
                            objectUri.getStore()
                                .createRevisionContent(objectUri,
                                                       revisionDescriptor,
                                                       revisionContent);
                        }
                    }
                    
                    objectUri.getStore()
                        .storeRevisionDescriptor
                        (objectUri, revisionDescriptor);
                    
                } catch (RevisionDescriptorNotFoundException e) {
                    // Should NEVER happen.
                    // Basically, it would mean that there is no initial
                    // revision, which is incorrect since the object
                    // HAS revisions.
                    
                    // Invoke interceptors
                    invokeInterceptors(token, revisionDescriptors,
                                       revisionDescriptor,
                                       revisionContent, PRE_STORE);
                    
                    objectUri.getStore()
                        .createRevisionDescriptor(objectUri,
                                                  revisionDescriptor);
                }
            }
            // Updating the descriptors object
            revisionDescriptors
                .setSuccessors(newRevisionNumber, new Vector());
            revisionDescriptors.setLatestRevision(newRevisionNumber);
        }
        
        // We now store the updated revision descriptors
        try {
            objectUri.getStore()
                .storeRevisionDescriptors(objectUri, revisionDescriptors);
        } catch (RevisionDescriptorNotFoundException e) {
            // Problem ...
            e.printStackTrace();
        }
        
        // Invoke interceptors
        invokeInterceptors(token, revisionDescriptors, revisionDescriptor,
                           revisionContent, POST_STORE);
        
    }
    
    
    /**
     * Create new revision based on a previous revision.
     *
     * @param strUri Uri
     * @param branch Branch in which to create the revision
     * @param newRevisionDescriptor New revision descriptor
     * @param revisionContent Node revision content
     */
    public void create(SlideToken token, String strUri, String branch,
                       NodeRevisionDescriptor newRevisionDescriptor,
                       NodeRevisionContent revisionContent)
        throws ObjectNotFoundException, AccessDeniedException,
        RevisionAlreadyExistException, LinkedObjectNotFoundException,
        ServiceAccessException, RevisionDescriptorNotFoundException,
        ObjectLockedException, NodeNotVersionedException,
        BranchNotFoundException {
        
        Uri objectUri = namespace.getUri(token, strUri);
        
        NodeRevisionDescriptors revisionDescriptors = objectUri.getStore()
            .retrieveRevisionDescriptors(objectUri);
        
        // Retrieving latest revision numbers
        NodeRevisionNumber branchLatestRevisionNumber =
            revisionDescriptors.getLatestRevision(branch);
        
        if (branchLatestRevisionNumber == null) {
            throw new BranchNotFoundException(strUri, branch);
        }
        
        create(token, strUri, branchLatestRevisionNumber,
               newRevisionDescriptor, revisionContent);
        
    }
    
    
    /**
     * Create a branch based on specified revision.
     *
     * @param strUri Uri
     * @param branchName Name of the new branch
     * @param basedOnRevisionDescriptor Node revision descriptor of the
     * revision on which the new branch is based on
     */
    public void fork(SlideToken token, String strUri, String branchName,
                     NodeRevisionDescriptor basedOnRevisionDescriptor)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionDescriptorNotFoundException, ObjectLockedException,
        NodeNotVersionedException, RevisionAlreadyExistException {
        
        fork(token, strUri, branchName,
             basedOnRevisionDescriptor.getRevisionNumber());
        
    }
    
    
    /**
     * Create a branch based on specified revision.
     *
     * @param strUri Uri
     * @param branchName Name of the new branch
     * @param basedOnRevisionNumber Node revision number of
     * the revision on which the new branch is based on
     */
    public void fork(SlideToken token, String strUri, String branchName,
                     NodeRevisionNumber basedOnRevisionNumber)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionDescriptorNotFoundException, ObjectLockedException,
        NodeNotVersionedException, RevisionAlreadyExistException {
        
        if (branchName.equals(NodeRevisionDescriptors.MAIN_BRANCH))
            return;
        
        // Retrieve the associated object
        ObjectNode associatedObject = structureHelper.retrieve(token, strUri);
        
        // Next we do a security check and a locking check for modifyRevisions
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getCreateRevisionMetadataAction());
        lockHelper.checkLock
            (token, associatedObject,
             namespaceConfig.getCreateRevisionMetadataAction());
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getCreateRevisionContentAction());
        lockHelper.checkLock(token, associatedObject,
                             namespaceConfig.getCreateRevisionContentAction());
        
        Uri objectUri = namespace.getUri(token, strUri);
        
        // Retrieve the revision table
        NodeRevisionDescriptors revisionDescriptors =
            objectUri.getStore()
            .retrieveRevisionDescriptors(objectUri);
        
        if (!revisionDescriptors.isVersioned()) {
            // Invalid function call : we try to create a revision, but the
            // descriptors won't allow it
            throw new NodeNotVersionedException(strUri);
        }
        
        // Retrieving the revision on which the new branch is based.
        NodeRevisionDescriptor basedOnRevisionDescriptor =
            objectUri.getStore().retrieveRevisionDescriptor
            (objectUri, basedOnRevisionNumber);
        NodeRevisionContent basedOnRevisionContent = null;
        try {
            basedOnRevisionContent =
                objectUri.getStore().retrieveRevisionContent
                (objectUri,
                 basedOnRevisionDescriptor);
        } catch (RevisionNotFoundException e) {
        }
        
        // Create a revision number suited for the new branch
        NodeRevisionNumber branchedRevisionNumber =
            new NodeRevisionNumber(basedOnRevisionNumber, true);
        
        basedOnRevisionDescriptor.setRevisionNumber(branchedRevisionNumber);
        basedOnRevisionDescriptor.setBranchName(branchName);
        
        revisionDescriptors.setUri(strUri);
        revisionDescriptors.setLatestRevision
            (branchName, branchedRevisionNumber);
        revisionDescriptors.addSuccessor
            (basedOnRevisionNumber, branchedRevisionNumber);
        
        // Invoke interceptors
        invokeInterceptors(token, revisionDescriptors,
                           basedOnRevisionDescriptor,
                           basedOnRevisionContent, PRE_STORE);
        
        // Storing back everything
        if (basedOnRevisionContent != null) {
            objectUri.getStore().createRevisionContent
                (objectUri, basedOnRevisionDescriptor, basedOnRevisionContent);
        }
        objectUri.getStore().createRevisionDescriptor
            (objectUri, basedOnRevisionDescriptor);
        objectUri.getStore().storeRevisionDescriptors
            (objectUri, revisionDescriptors);
        
        // Invoke interceptors
        invokeInterceptors(token, revisionDescriptors,
                           basedOnRevisionDescriptor,
                           basedOnRevisionContent, POST_STORE);
        
    }
    
    
    /**
     * Merge specified branches into a single branch.
     *
     * @param strUri Uri
     * @param mainBranch Branch into which the other branch will be merged
     * @param branch Branch to merge into main branch
     * @param newRevisionDescriptor New revision descriptor
     * @param revisionContent Node revision content
     */
    public void merge(SlideToken token, String strUri,
                      NodeRevisionDescriptor mainBranch,
                      NodeRevisionDescriptor branch,
                      NodeRevisionDescriptor newRevisionDescriptor,
                      NodeRevisionContent revisionContent)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionDescriptorNotFoundException, ObjectLockedException,
        NodeNotVersionedException, BranchNotFoundException,
        RevisionAlreadyExistException {
        
        merge(token, strUri, mainBranch.getBranchName(),
              branch.getBranchName(), newRevisionDescriptor, revisionContent);
        
    }
    
    
    /**
     * Merge specified branches into a single branch.
     *
     * @param strUri Uri
     * @param mainBranch Branch into which the other branch will be merged
     * @param branch Branch to merge into main branch
     * @param newRevisionDescriptor New revision descriptor
     * @param revisionContent Node revision content
     */
    public void merge(SlideToken token, String strUri,
                      String mainBranch, String branch,
                      NodeRevisionDescriptor newRevisionDescriptor,
                      NodeRevisionContent revisionContent)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionDescriptorNotFoundException, ObjectLockedException,
        NodeNotVersionedException, BranchNotFoundException,
        RevisionAlreadyExistException {
        
        // Retrieve the associated object
        ObjectNode associatedObject = structureHelper.retrieve(token, strUri);
        
        // Next we do a security check and a locking check for modifyRevisions
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getCreateRevisionMetadataAction());
        lockHelper.checkLock
            (token, associatedObject,
             namespaceConfig.getCreateRevisionMetadataAction());
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getCreateRevisionContentAction());
        lockHelper.checkLock(token, associatedObject,
                             namespaceConfig.getCreateRevisionContentAction());
        
        setDefaultProperties(associatedObject, newRevisionDescriptor);
        
        Uri objectUri = namespace.getUri(token, strUri);
        
        // Retrieve the revision table
        NodeRevisionDescriptors revisionDescriptors =
            objectUri.getStore().retrieveRevisionDescriptors(objectUri);
        
        if (!revisionDescriptors.isVersioned()) {
            // Invalid function call : we try to create a revision, but the
            // descriptors won't allow it
            throw new NodeNotVersionedException(strUri);
        }
        
        // Retrieving latest revision numbers
        NodeRevisionNumber mainBranchLatestRevisionNumber =
            revisionDescriptors.getLatestRevision(mainBranch);
        NodeRevisionNumber branchLatestRevisionNumber =
            revisionDescriptors.getLatestRevision(branch);
        
        if (mainBranchLatestRevisionNumber == null) {
            throw new BranchNotFoundException(strUri, mainBranch);
        }
        if (branchLatestRevisionNumber == null) {
            throw new BranchNotFoundException(strUri, branch);
        }
        
        NodeRevisionNumber newRevisionNumber =
            new NodeRevisionNumber(mainBranchLatestRevisionNumber);
        
        newRevisionDescriptor.setRevisionNumber(newRevisionNumber);
        newRevisionDescriptor.setBranchName(branch);
        
        revisionDescriptors.addSuccessor
            (mainBranchLatestRevisionNumber, newRevisionNumber);
        revisionDescriptors.addSuccessor
            (branchLatestRevisionNumber, newRevisionNumber);
        revisionDescriptors.setLatestRevision(mainBranch, newRevisionNumber);
        
        // Invoke interceptors
        invokeInterceptors(token, revisionDescriptors, newRevisionDescriptor,
                           revisionContent, PRE_STORE);
        
        // Storing back everything
        if (revisionContent != null) {
            objectUri.getStore().createRevisionContent
                (objectUri, newRevisionDescriptor, revisionContent);
        }
        objectUri.getStore().createRevisionDescriptor
            (objectUri, newRevisionDescriptor);
        objectUri.getStore().storeRevisionDescriptors
            (objectUri, revisionDescriptors);
        
        // Invoke interceptors
        invokeInterceptors(token, revisionDescriptors, newRevisionDescriptor,
                           revisionContent, POST_STORE);
        
    }
    
    
    /**
     * Update contents of an existing revision.
     *
     * @param strUri Uri
     * @param revisionDescriptor Revision descriptor
     * @param revisionContent Revision content
     */
    public void store(SlideToken token, String strUri,
                      NodeRevisionDescriptor revisionDescriptor,
                      NodeRevisionContent revisionContent)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionDescriptorNotFoundException, ObjectLockedException,
        RevisionNotFoundException {
        
        // Retrieve the associated object
        ObjectNode associatedObject = structureHelper.retrieve(token, strUri);
        
        // Next we do a security check and a locking check for modifyRevisions
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getModifyRevisionMetadataAction());
        lockHelper.checkLock
            (token, associatedObject,
             namespaceConfig.getModifyRevisionMetadataAction());
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getModifyRevisionContentAction());
        lockHelper.checkLock(token, associatedObject,
                             namespaceConfig.getModifyRevisionContentAction());
        
        setDefaultProperties(associatedObject, revisionDescriptor);
        
        Uri objectUri = namespace.getUri(token, strUri);
        
        // Retrieve the revision table
        NodeRevisionDescriptors revisionDescriptors =
            objectUri.getStore().retrieveRevisionDescriptors(objectUri);
        NodeRevisionDescriptor oldRevisionDescriptor =
            objectUri.getStore().retrieveRevisionDescriptor
            (objectUri, revisionDescriptor.getRevisionNumber());
        
        // Invoke interceptors
        invokeInterceptors(token, revisionDescriptors, revisionDescriptor,
                           revisionContent, PRE_STORE);
        
        if (revisionContent != null) {
            try {
                objectUri.getStore().retrieveRevisionContent
                    (objectUri, revisionDescriptor);
                objectUri.getStore().storeRevisionContent
                    (objectUri, revisionDescriptor, revisionContent);
            } catch (RevisionNotFoundException e) {
                try {
                    objectUri.getStore().createRevisionContent
                        (objectUri, revisionDescriptor, revisionContent);
                } catch (RevisionAlreadyExistException ex) {
                    // Should never happen
                    ex.printStackTrace();
                }
            }
        }
        objectUri.getStore().storeRevisionDescriptor
            (objectUri, revisionDescriptor);
        
        // Invoke interceptors
        invokeInterceptors(token, revisionDescriptors, revisionDescriptor,
                           revisionContent, POST_STORE);
        
    }
    
    
    /**
     * Remove all revisions at this Uri.
     *
     * @param revisionDescriptors Node revision descriptors
     */
    public void remove(SlideToken token,
                       NodeRevisionDescriptors revisionDescriptors)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionDescriptorNotFoundException, ObjectLockedException {
        
        // Retrieve the associated object
        ObjectNode associatedObject =
            structureHelper.retrieve(token, revisionDescriptors.getUri());
        
        // Next we do a security check and a locking check for modifyRevisions
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getRemoveRevisionMetadataAction());
        lockHelper.checkLock
            (token, associatedObject,
             namespaceConfig.getRemoveRevisionMetadataAction());
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getRemoveRevisionContentAction());
        lockHelper.checkLock(token, associatedObject,
                             namespaceConfig.getRemoveRevisionContentAction());
        
        Uri objectUri = namespace.getUri(token, revisionDescriptors.getUri());
        
        objectUri.getStore().removeRevisionDescriptors(objectUri);
        
    }
    
    
    /**
     * Remove specified revision.
     *
     * @param strUri Uri
     * @param revisionDescriptor Node revision descriptor
     */
    public void remove(SlideToken token, String strUri,
                       NodeRevisionDescriptor revisionDescriptor)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionDescriptorNotFoundException, ObjectLockedException {
        
        remove(token, strUri, revisionDescriptor.getRevisionNumber());
        
    }
    
    
    /**
     * Remove specified revision.
     *
     * @param strUri Uri
     * @param revisionNumber Revision number
     */
    public void remove(SlideToken token, String strUri,
                       NodeRevisionNumber revisionNumber)
        throws ObjectNotFoundException, AccessDeniedException,
        LinkedObjectNotFoundException, ServiceAccessException,
        RevisionDescriptorNotFoundException, ObjectLockedException {
        
        // Retrieve the associated object
        ObjectNode associatedObject = structureHelper.retrieve(token, strUri);
        
        // Next we do a security check and a locking check for modifyRevisions
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getRemoveRevisionMetadataAction());
        lockHelper.checkLock
            (token, associatedObject,
             namespaceConfig.getRemoveRevisionMetadataAction());
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getRemoveRevisionContentAction());
        lockHelper.checkLock(token, associatedObject,
                             namespaceConfig.getRemoveRevisionContentAction());
        
        Uri objectUri = namespace.getUri(token, strUri);
        
        NodeRevisionDescriptor revisionDescriptor =
            objectUri.getStore().retrieveRevisionDescriptor
            (objectUri, revisionNumber);
        
        objectUri.getStore().removeRevisionContent
            (objectUri, revisionDescriptor);
        objectUri.getStore()
            .removeRevisionDescriptor(objectUri, revisionNumber);
        
    }
    
    
    // ------------------------------------------------------ Protected Methods
    
    
    /**
     * Create new revision based on a previous revision.
     *
     * @param strUri Uri
     * @param oldRevisionDescriptor Node revision descriptor on which the
     * new revision is based
     * @param newRevisionDescriptor New revision descriptor
     * @param revisionContent Node revision content
     */
    protected void create(SlideToken token, String strUri,
                          NodeRevisionNumber basedOnRevisionNumber,
                          NodeRevisionDescriptor newRevisionDescriptor,
                          NodeRevisionContent revisionContent)
        throws ObjectNotFoundException, AccessDeniedException,
        RevisionAlreadyExistException, LinkedObjectNotFoundException,
        ServiceAccessException, RevisionDescriptorNotFoundException,
        ObjectLockedException, NodeNotVersionedException {
        
        // Retrieve the associated object
        ObjectNode associatedObject = structureHelper.retrieve(token, strUri);
        
        // Next we do a security check and a locking check for modifyRevisions
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getCreateRevisionMetadataAction());
        lockHelper.checkLock
            (token, associatedObject,
             namespaceConfig.getCreateRevisionMetadataAction());
        securityHelper.checkCredentials
            (token, associatedObject,
             namespaceConfig.getCreateRevisionContentAction());
        lockHelper.checkLock(token, associatedObject,
                             namespaceConfig.getCreateRevisionContentAction());
        
        setDefaultProperties(associatedObject, newRevisionDescriptor);
        
        Uri objectUri = namespace.getUri(token, strUri);
        
        // Retrieve the revision table
        NodeRevisionDescriptors revisionDescriptors =
            objectUri.getStore()
            .retrieveRevisionDescriptors(objectUri);
        
        if (!revisionDescriptors.isVersioned()) {
            // Invalid function call : we try to create a revision, but the
            // descriptors won't allow it
            throw new NodeNotVersionedException(strUri);
        }
        
        // Retrieve the old revision descriptor, just to make sure that the old
        //  revision we base the new one upon really exists
        NodeRevisionDescriptor realOldRevisionDescriptor =
            objectUri.getStore().retrieveRevisionDescriptor
            (objectUri, basedOnRevisionNumber);
        
        // We check that the old revision doesn't have successors, that is we :
        // - check to see if it's the latest revision in a branch
        // - store that information for later use
        NodeRevisionNumber latestNumberInBranch =
            revisionDescriptors.getLatestRevision
            (realOldRevisionDescriptor.getBranchName());
        if (!realOldRevisionDescriptor.getRevisionNumber()
            .equals(latestNumberInBranch)) {
            throw new RevisionAlreadyExistException
                (objectUri.toString(), new NodeRevisionNumber
                     (basedOnRevisionNumber));
        }
        
        // Next, generate the new revision's number
        newRevisionDescriptor.setRevisionNumber
            (new NodeRevisionNumber(basedOnRevisionNumber));
        // Set the creation date
        newRevisionDescriptor.setCreationDate(new Date());
        
        // Initialize the branch name in the new descriptor
        newRevisionDescriptor.setBranchName
            (realOldRevisionDescriptor.getBranchName());
        
        // Invoke interceptors
        invokeInterceptors(token, revisionDescriptors, newRevisionDescriptor,
                           revisionContent, PRE_STORE);
        
        // Update the revision graph in the revision descriptors
        revisionDescriptors
            .addSuccessor(revisionDescriptors.getLatestRevision(),
                          newRevisionDescriptor.getRevisionNumber());
        revisionDescriptors
            .setLatestRevision(newRevisionDescriptor.getBranchName(),
                               newRevisionDescriptor.getRevisionNumber());
        if (revisionContent != null) {
            // Storing the new revision contents
            objectUri.getStore()
                .createRevisionContent(objectUri, newRevisionDescriptor,
                                       revisionContent);
        }
        // Now creating the revision desriptor in the store
        objectUri.getStore()
            .createRevisionDescriptor(objectUri, newRevisionDescriptor);
        
        // We now store the updated revision descriptors
        try {
            objectUri.getStore()
                .storeRevisionDescriptors(objectUri, revisionDescriptors);
        } catch (RevisionDescriptorNotFoundException e) {
            // Problem ...
            e.printStackTrace();
        }
        
        // Invoke interceptors
        invokeInterceptors(token, revisionDescriptors, newRevisionDescriptor,
                           revisionContent, POST_STORE);
        
    }
    
    
    /**
     * Set default properties for a revision descriptors.
     */
    protected void setDefaultProperties
        (ObjectNode associatedObject,
         NodeRevisionDescriptor revisionDescriptor) {
        // Retrieving the roles of the associated object
        Enumeration roles = securityHelper.getRoles(associatedObject);
        while (roles.hasMoreElements()) {
            String role = (String) roles.nextElement();
            Enumeration defaultProperties =
                namespaceConfig.getDefaultProperties(role);
            revisionDescriptor.setDefaultProperties(defaultProperties);
        }
    }
    
    
    /**
     * Invoke content interceptors.
     */
    protected void invokeInterceptors
        (SlideToken token, NodeRevisionDescriptors revisionDescriptors,
         NodeRevisionDescriptor revisionDescriptor,
         NodeRevisionContent revisionContent, int type) {
        ContentInterceptor[] contentInterceptors =
            namespace.getContentInterceptors();
        for (int i = 0; i < contentInterceptors.length; i++) {
            switch (type) {
                case PRE_STORE:
                    contentInterceptors[i].preStoreContent
                        (token, revisionDescriptors,
                         revisionDescriptor, revisionContent);
                    break;
                case POST_STORE:
                    contentInterceptors[i].postStoreContent
                        (token, revisionDescriptors,
                         revisionDescriptor, revisionContent);
                    break;
                case POST_RETRIEVE:
                    contentInterceptors[i].postRetrieveContent
                        (token, revisionDescriptors,
                         revisionDescriptor, revisionContent);
                    break;
            }
        }
    }
    
    
}
