/*
 * $Header: /home/cvspublic/jakarta-slide/src/webdav/client/src/org/apache/webdav/lib/WebdavSession.java,v 1.12 2001/05/16 16:51:38 jericho Exp $
 * $Revision: 1.12 $
 * $Date: 2001/05/16 16:51:38 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */

package org.apache.webdav.lib;

import java.io.IOException;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.net.MalformedURLException;
import org.apache.util.HttpURL;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.Credentials;
import org.apache.commons.httpclient.ConnectionInterceptor;
import org.apache.commons.httpclient.StreamInterceptor;

/**
 *
 * This WebdavSession class is for the multi-session management of
 * WebDAV clients.   This class saves and restores the requested client.
 *
 * Clients that use persistent connections SHOULD limit the number of
 * simultaneous connections that they maintain to a given server. A
 * single-user client SHOULD NOT maintain more than 2 connections with
 * any server or proxy. A proxy SHOULD use up to 2*N connections to
 * another server or proxy, where N is the number of simultaneously
 * active users. These guidelines are intended to improve HTTP response
 * times and avoid congestion.
 *
 * @author <a href="mailto:jericho@thinkfree.com">Park, Sung-Gu</a>
 */
public class WebdavSession implements ConnectionInterceptor {

    // ---------------------------------------------------- Instance Variables


    /**
     * Infomation hashtable for WebDAV clients  HTTP/1.1 communication.
     */
    private static Hashtable clientInfo = new Hashtable();
    
    
    /**
     * The Http client instance.
     */
    protected HttpClient client;


    /**
     * Debug level.
     */
    protected int debug = 0;


    // ------------------------------------------------------------- Properties


    /**
     * Set debug level.
     */
    public void setDebug(int debug) {
        this.debug = debug;
        if (client != null) {
            client.setDebug(debug);
        }
    }


    // ------------------------------------------------------  Public methods


    /**
     * Get a <code>HttpClient</code> instance.
     * This method returns a new client instance for the first time.
     * And it is saved util it's closed or reset.
     *
     * @param httpURL The http URL to connect.  only used the authority part.
     * @return An instance of <code>HttpClient</code>.
     * @exception IOException
     */
    public HttpClient getSessionInstance(HttpURL httpURL)
        throws IOException {

        return getSessionInstance(httpURL, false);
    }


    /**
     * Get a <code>HttpClient</code> instance.
     * This method returns a new client instance, when reset is true.
     *
     * @param httpURL The http URL to connect.  only used the authority part.
     * @param reset The reset flag to represent whether the saved information
     *              is used or not.
     * @return An instance of <code>HttpClient</code>.
     * @exception IOException
     */
    public synchronized HttpClient getSessionInstance(HttpURL httpURL,
                                                      boolean reset)
        throws IOException {
        
        String authority = httpURL.getAuthority();
        if (reset)
            clientInfo.remove(authority);
        HttpClient client = (HttpClient) clientInfo.get(authority);
        if (client == null) {
            client = new HttpClient();
            client.startSession(httpURL.getHost(), httpURL.getPort());
            String userName = httpURL.getUserName();
            if (userName != null && userName.length() > 0) {
                String password = httpURL.getPassword();
                client.setCredentials(new Credentials(userName, password));
            }
            client.setConnectionInterceptor(this);
            clientInfo.put(authority, client);
        }
        client.setDebug(debug);        

        return client;
    }

   
    /**
     * Set the session by the given client.
     * In order to save infomration of the client, it must be used.
     *
     * @param client The Http client instance.
     * @exception IOException
     */
    protected synchronized void setSession(HttpClient client)
        throws IOException {
        
        HttpURL httpURL =
            new HttpURL(client.getUserName(), client.getPassword(),
                        client.getHost(), client.getPort());
        clientInfo.put(httpURL.getAuthority(), client);
    }


    /**
     * Test a session to be connected.
     *
     * @param httpURL The http URL to connect. only used the authority part.
     * @return true if connected aleady.
     * @exception MalformedURLException
     */
    public boolean isSession(HttpURL httpURL)
        throws MalformedURLException {
        
        return isSession(httpURL.getAuthority());
    }


    /**
     * Test a session to be connected.
     *
     * @param authority The authority string.
     * @return true if connected aleady.
     */
    public synchronized boolean isSession(String authority) {
        
        Enumeration authorities = clientInfo.keys();
        while (authorities.hasMoreElements()) {
            if (authority.equals((String) authorities.nextElement())) {
                return true;
            }
        }
        return false;
    }
    
    
    /**
     * Find the sessions having the part of the authority string.
     * The <code>partOfAuthority</code> can be user or host part.
     *
     * @param partOfAuthority The part of the authority string.
     * @return Authorities to include the given string.
     */
    public synchronized Enumeration findSessions(String partOfAuthority) {
        
        Vector sessions = new Vector();
        Enumeration authorities = clientInfo.keys();
        while (authorities.hasMoreElements()) {
            String authority = (String) authorities.nextElement();
            if (authority.indexOf(partOfAuthority) >= 0) {
                sessions.addElement(authority);
            }
        }
        
        return sessions.elements();
    }


    /**
     * Get all session information.
     *
     * @return Authorities of all session.
     */
    public synchronized Enumeration getSessions() {
        return clientInfo.keys();
    }


    /**
     * Close an session and delete the connection information.
     *
     * @param client The HttpClient instance.
     * @exception IOException Error in closing socket.
     */
    public synchronized void closeSession(HttpClient client)
        throws IOException {

        HttpURL httpURL =
            new HttpURL(client.getUserName(), client.getPassword(),
                        client.getHost(), client.getPort());
        client.endSession();
        clientInfo.remove(httpURL.getAuthority());
        // Remove the progress listener.
        // webdavClient.getProgressUtil().addProgressListener(this);
    }


    /**
     * Connect.
     */
    public void connect() {
        if (debug == 9) {
            System.err.println("CLIENT: connected.");
        }
    }


    /**
     * Disconnect.
     */
    public void disconnect() {
        if (debug == 9) {
            System.err.println("CLIENT: disconnected.");
        }
    }


    /**
     * Retry.
     * 
     * @return boolean true if a retry should be attempted
     */
    public boolean retry(int status) {
        if (debug == 9) {
            System.err.println("CLIENT: retrying.");
        }
        return false;
    }


    /**
     * Recieved an informational status code.
     * 
     * @return boolean true if a retry should be attempted
     */
    public boolean info(int status, Hashtable headers) {
        if (debug == 9) {
            System.err.println("CLIENT: Status: " + status + "\nHeaders: ");
            Enumeration keys = headers.keys();
            while (keys.hasMoreElements()) {
                String header = (String) keys.nextElement();
                String value = (String) headers.get(header);
                System.err.println("\t" + header + ": " + value);
            }
        }
        return false;
    }


    /**
     * Unexpected error.
     * 
     * @param status Status code; can be equal to -1 if status code is not 
     * known
     * @param e Underlying exception; can be null
     * @return boolean true if processing of the request should be stopped
     */
    public boolean error(int status, Exception e) {
        if (debug == 9) {
            System.err.println("CLIENT: Status: " + status + ", Exception: " +
                               e.getMessage());
        }
        return false;
    }


    /**
     * Sent request.
     */
    public void sentRequest() {
        if (debug == 9) {
            System.err.println("CLIENT: sent request.");
        }
    }


    /**
     * Recieved response.
     */
    public void recievedResponse() {
        if (debug == 9) {
            System.err.println("CLIENT: recieved response.");
        }
    }


    /**
     * Recieved expectation.
     */
    public void recievedExpectation() {
        if (debug == 9) {
            System.err.println("CLIENT: recieved expectation.");
        }
    }


    /**
     * Authenticate.
     */
    public void requiredAuthentication() {
        if (debug == 9) {
            System.err.println("CLIENT: required authentication.");
        }
    }


    /**
     * Authenticate.
     */
    public void authenticate() {
        if (debug == 9) {
            System.err.println("CLIENT: authenticating.");
        }
    }


    /**
     * Progressing by the progress event.
     *
     * @param pe The progress event.
     */
    /*
    public void progressing(ProgressEvent pe) {
        if (debug > 3)
            System.err.println("[EVENT/WebdavSession] " +
                               "action:" + pe.getAction() +
                               ((pe.getResourceName() == null) ? "" :
                               ", resource:" + pe.getResourceName()) +
                               ", soMany:" + pe.getSoMany() +
                               ", remained:" + pe.getRemainedSize() +
                               ", total:" + pe.getTotalSize());
        getProgressUtil().fireProgress(pe);
    }
    */


    /**
     * Get the utility of this progress event and listener.
     *
     * @return ProgressUtil
     */
    /*
    public ProgressUtil getProgressUtil() {
        return sessionProgress;
    }
    */
}
