/*
 * $Header: /home/cvspublic/jakarta-slide/src/webdav/server/org/apache/slide/webdav/method/GetMethod.java,v 1.10 2001/06/13 04:34:09 remm Exp $
 * $Revision: 1.10 $
 * $Date: 2001/06/13 04:34:09 $
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Tomcat", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * [Additional notices, if required by prior licensing conditions]
 *
 */

package org.apache.slide.webdav.method;

import java.security.Principal;
import java.io.*;
import java.util.*;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import javax.servlet.*;
import javax.servlet.http.*;
import org.apache.util.WebdavStatus;
import org.apache.slide.common.*;
import org.apache.slide.security.AccessDeniedException;
import org.apache.slide.webdav.*;
import org.apache.slide.util.Messages;
import org.apache.slide.content.*;
import org.apache.slide.lock.*;
import org.apache.slide.security.*;
import org.apache.slide.structure.*;

/**
 * GET method.
 *
 * @author <a href="mailto:remm@apache.org">Remy Maucherat</a>
 */
public class GetMethod extends WebdavMethod {
    
    
    // -------------------------------------------------------------- Constants
    
    
    protected final int BUFFER_SIZE = 2048;
    
    
    /**
     * The set of SimpleDateFormat formats to use in getDateHeader().
     */
    protected static final SimpleDateFormat formats[] = {
        new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz", Locale.US),
        new SimpleDateFormat("EEEEEE, dd-MMM-yy HH:mm:ss zzz", Locale.US),
        new SimpleDateFormat("EEE MMMM d HH:mm:ss yyyy", Locale.US)
    };
    
    
    /**
     * MIME multipart separation string
     */
    protected static final String mimeSeparation = "SLIDE_MIME_BOUNDARY";
    
    
    /**
     * The input buffer size to use when serving resources.
     */
    protected int input = 2048;
    
    
    /**
     * The output buffer size to use when serving resources.
     */
    protected int output = 2048;
    
    
    /**
     * Print content.
     */
    protected boolean printContent = true;
    
    
    // ----------------------------------------------------- Instance Variables
    
    
    /**
     * Resource to be retrieved.
     */
    private String resourcePath;
    
    
    // ----------------------------------------------------------- Constructors
    
    
    /**
     * GET Method constructor.
     *
     * @param token Namespace access token
     * @param req HTTP request
     * @param resp HTTP response
     */
    public GetMethod(GenericServlet servlet, NamespaceAccessToken token,
                     HttpServletRequest req, HttpServletResponse resp) {
        super(servlet, token, req, resp);
    }
    
    
    // ------------------------------------------------------ Protected Methods
    
    
    /**
     * Parse XML request.
     */
    protected void parseRequest()
        throws WebdavException {
        resourcePath = requestUri;
        if (resourcePath == null) {
            resourcePath = "/";
        }
    }
    
    
    /**
     * Execute request.
     *
     * @exception WebdavException Can't access resource
     */
    protected void executeRequest()
        throws WebdavException {
        
        try {
            
            // Then we must get object contents ...
            
            ObjectNode object = structure.retrieve(slideToken, resourcePath);
            NodeRevisionDescriptors revisionDescriptors =
                content.retrieve(slideToken, resourcePath);
            
            if (revisionDescriptors.hasRevisions()) {
                
                // Retrieve latest revision descriptor
                NodeRevisionDescriptor revisionDescriptor =
                    content.retrieve(slideToken, revisionDescriptors);
                
                if (isCollection(revisionDescriptor)) {
                    
                    Writer writer = resp.getWriter();
                    resp.setContentType("text/html");
                    displayDirectoryBrowsing(object, writer);
                    writer.close();
                    
                } else {
                    
                    ResourceInfo resourceInfo =
                        new ResourceInfo(resourcePath, revisionDescriptor);
                    
                    // Checking If headers
                    if (!checkIfHeaders(req, resp, resourceInfo))
                        return;
                    
                    ServletOutputStream os = resp.getOutputStream();
                    InputStream         is = null;
                    
                    if (printContent) {
                        is = content.retrieve
                            (slideToken, revisionDescriptors,
                             revisionDescriptor).streamContent();
                    }
                    
                    Vector ranges = parseRange(req, resp, resourceInfo);
                    
                    // ETag header
                    resp.setHeader("ETag", getETag(resourceInfo, true));
                    resp.addHeader
                        ("Last-Modified",
                         revisionDescriptor.getCreationDate().toString());
                    
                    if ( ((ranges == null) || (ranges.isEmpty()))
                         && (req.getHeader("Range") == null) ) {
                        
                        resp.setContentType
                            (revisionDescriptor.getContentType());
                        resp.setContentLength
                            ((int) revisionDescriptor.getContentLength());
                        
                        // Copy the input stream to our output stream
                        // (if requested)
                        if (printContent) {
                            resp.setBufferSize(output);
                            copy(resourceInfo, is, os);
                        }
                        
                    } else {
                        
                        if ((ranges == null) || (ranges.isEmpty()))
                            return;
                        
                        // Partial content response.
                        
                        resp.setStatus(HttpServletResponse.SC_PARTIAL_CONTENT);
                        
                        if (ranges.size() == 1) {
                            
                            Range range = (Range) ranges.elementAt(0);
                            resp.addHeader("Content-Range", "bytes "
                                           + range.start
                                           + "-" + range.end + "/"
                                           + range.length);
                            resp.setContentLength((int) range.length);
                            resp.setContentType
                                (revisionDescriptor.getContentType());
                            
                            if (printContent) {
                                resp.setBufferSize(output);
                                copy(resourceInfo, is, os, range);
                            }
                            
                        } else {
                            
                            resp.setContentType
                                ("multipart/byteranges; boundary="
                                 + mimeSeparation);
                            
                            if (printContent) {
                                resp.setBufferSize(output);
                                copy(resourceInfo, is, os,
                                     ranges.elements(),
                                     revisionDescriptor.getContentType());
                            }
                            
                        }
                        
                    }
                    
                }
                
            } else {
                
                Writer writer = resp.getWriter();
                resp.setContentType("text/html");
                displayDirectoryBrowsing(object, writer);
                writer.close();
                
            }
            
        } catch (RevisionNotFoundException e) {
            // 404 - Not found
            System.out.println(e.getMessage());
            resp.setStatus(WebdavStatus.SC_NOT_FOUND);
    } catch (RevisionContentNotFoundException e) {
            // 404 - Not found
            System.out.println(e.getMessage());
            resp.setStatus(WebdavStatus.SC_NOT_FOUND);
    } catch (ObjectLockedException e) {
            // Locked
            System.out.println(e.getMessage());
            resp.setStatus(WebdavStatus.SC_LOCKED);
    } catch (RevisionDescriptorNotFoundException e) {
            // 404 - Not found
            System.out.println(e.getMessage());
            resp.setStatus(WebdavStatus.SC_NOT_FOUND);
    } catch (ObjectNotFoundException e) {
            // 404 - Not found
            resp.setStatus(WebdavStatus.SC_NOT_FOUND);
    } catch (LinkedObjectNotFoundException e) {
            // 404 - Not found
            System.out.println(e.getMessage());
            resp.setStatus(WebdavStatus.SC_NOT_FOUND);
    } catch (AccessDeniedException e) {
            // 403 - Forbidden
            System.out.println(e.getMessage());
            resp.setStatus(WebdavStatus.SC_FORBIDDEN);
    } catch (ServiceAccessException e) {
        // 500 - Internal server error
            System.out.println(e.getMessage());
            resp.setStatus(WebdavStatus.SC_INTERNAL_SERVER_ERROR);
        throw new WebdavException(WebdavStatus.SC_INTERNAL_SERVER_ERROR);
    } catch (IOException e) {
            e.printStackTrace();
            // Unable to write to output stream.
        // 500 - Internal server error
            resp.setStatus(WebdavStatus.SC_INTERNAL_SERVER_ERROR);
        } catch (RuntimeException e) {
            e.printStackTrace();
        // 500 - Internal server error
            resp.setStatus(WebdavStatus.SC_INTERNAL_SERVER_ERROR);
    }
        
    }
    
    
    /**
     * Display a directory browsing page.
     */
    protected void displayDirectoryBrowsing(ObjectNode object,
                                            Writer servletWriter)
        throws IOException {
        
        String contextPath = req.getContextPath();
        if (contextPath == null)
            contextPath = "";
        
        String name = object.getUri();
        
        // Number of characters to trim from the beginnings of filenames
        int trim = name.length();
        if (!name.endsWith("/"))
            trim += 1;
        if (name.equals("/"))
            trim = 1;
        
        PrintWriter writer = new PrintWriter(servletWriter);
        
        // Render the page header
        writer.print("<html>\r\n");
        writer.print("<head>\r\n");
        writer.print("<title>");
        writer.print
            (Messages.format
             ("org.apache.slide.webdav.GetMethod.directorylistingfor", name));
        writer.print("</title>\r\n</head>\r\n");
        writer.print("<body bgcolor=\"white\">\r\n");
        writer.print("<table width=\"90%\" cellspacing=\"0\"" +
                     " cellpadding=\"5\" align=\"center\">\r\n");
        
        // Render the in-page title
        writer.print("<tr><td colspan=\"3\"><font size=\"+2\">\r\n<strong>");
        writer.print
            (Messages.format
             ("org.apache.slide.webdav.GetMethod.directorylistingfor", name));
        writer.print("</strong>\r\n</font></td></tr>\r\n");
        
        // Render the link to our parent (if required)
        String parentDirectory = name;
        if (parentDirectory.endsWith("/")) {
            parentDirectory =
                parentDirectory.substring(0, parentDirectory.length() - 1);
        }
        int slash = parentDirectory.lastIndexOf("/");
        if (slash >= 0) {
            String parent = name.substring(0, slash);
            writer.print("<tr><td colspan=\"5\" bgcolor=\"#ffffff\">\r\n");
            writer.print("<a href=\"");
            writer.print(URLEncode(contextPath));
            if (parent.equals(""))
                parent = "/";
            writer.print(parent);
            writer.print("\">");
            writer.print(Messages.format
                         ("org.apache.slide.webdav.GetMethod.parent", parent));
            writer.print("</a>\r\n");
            writer.print("</td></tr>\r\n");
        }
        
        Enumeration permissionsList = null;
        Enumeration locksList = null;
        
        try {
            
            permissionsList =
                security.enumeratePermissions(slideToken, object.getUri());
            locksList = lock.enumerateLocks(slideToken, object.getUri());
            
        } catch (SlideException e) {
            
            // Any security based exception will be trapped here
            
            // Any locking based exception will be trapped here
            
        }
        
        // Displaying ACL info
        displayPermissions(permissionsList, writer, false);
        
        // Displaying lock info
        displayLocks(locksList, writer, false);
        
        writer.print("<tr><td colspan=\"5\" bgcolor=\"#ffffff\">");
        writer.print("&nbsp;");
        writer.print("</td></tr>\r\n");
        
        // Render the column headings
        writer.print("<tr bgcolor=\"#cccccc\">\r\n");
        writer.print("<td align=\"left\" colspan=\"3\">");
        writer.print("<font size=\"+1\"><strong>");
        writer.print(Messages.message
                     ("org.apache.slide.webdav.GetMethod.filename"));
        writer.print("</strong></font></td>\r\n");
        writer.print("<td align=\"center\"><font size=\"+1\"><strong>");
        writer.print(Messages.message
                     ("org.apache.slide.webdav.GetMethod.size"));
        writer.print("</strong></font></td>\r\n");
        writer.print("<td align=\"right\"><font size=\"+1\"><strong>");
        writer.print(Messages.message
                     ("org.apache.slide.webdav.GetMethod.lastModified"));
        writer.print("</strong></font></td>\r\n");
        writer.print("</tr>\r\n");
        
        Enumeration resources = object.enumerateChildren();
        boolean shade = false;
        
        while (resources.hasMoreElements()) {
            
            String currentResource = (String) resources.nextElement();
            
            NodeRevisionDescriptor currentDescriptor = null;
            
            permissionsList = null;
            locksList = null;
            
            try {
                
                NodeRevisionDescriptors revisionDescriptors =
                    content.retrieve(slideToken, currentResource);
                // Retrieve latest revision descriptor
                currentDescriptor =
                    content.retrieve(slideToken, revisionDescriptors);
                
            } catch (SlideException e) {
                
                // Silent exception : Objects without any revision are
                // considered collections, and do not have any attributes
                
                // Any security based exception will be trapped here
                
                // Any locking based exception will be trapped here
                
            }
            
            try {
                
                permissionsList =
                    security.enumeratePermissions(slideToken, currentResource);
                locksList = lock.enumerateLocks(slideToken, currentResource);
                
            } catch (SlideException e) {
                
                // Any security based exception will be trapped here
                
                // Any locking based exception will be trapped here
                
            }
            
            String trimmed = currentResource.substring(trim);
            if (trimmed.equalsIgnoreCase("WEB-INF") ||
                trimmed.equalsIgnoreCase("META-INF"))
                continue;
            
            writer.print("<tr");
            if (shade) {
                writer.print(" bgcolor=\"dddddd\"");
            } else {
                writer.print(" bgcolor=\"eeeeee\"");
            }
            writer.print(">\r\n");
            shade = !shade;
            
            writer.print("<td align=\"left\" colspan=\"3\">&nbsp;&nbsp;\r\n");
            writer.print("<a href=\"");
            writer.print(URLEncode(contextPath + currentResource));
            writer.print("\"><tt>");
            writer.print(trimmed);
            if (isCollection(currentDescriptor)) {
                writer.print("/");
            }
            writer.print("</tt></a></td>\r\n");
            
            writer.print("<td align=\"right\"><tt>");
            if (isCollection(currentDescriptor))
                writer.print("&nbsp;");
            else
                writer.print(renderSize(currentDescriptor.getContentLength()));
            writer.print("</tt></td>\r\n");
            
            writer.print("<td align=\"right\"><tt>");
            if (currentDescriptor != null) {
                writer.print(currentDescriptor.getLastModified());
            } else {
                writer.print("&nbsp;");
            }
            writer.print("</tt></td>\r\n");
            
            writer.print("</tr>\r\n");
            
            // Displaying ACL info
            displayPermissions(permissionsList, writer, shade);
            
            // Displaying lock info
            displayLocks(locksList, writer, shade);
            
        }
        
        // Render the page footer
        writer.print("<tr><td colspan=\"5\">&nbsp;</td></tr>\r\n");
        writer.print("<tr><td colspan=\"3\" bgcolor=\"#cccccc\">");
        writer.print("<font size=\"-1\">");
        writer.print(Messages.message
                     ("org.apache.slide.webdav.GetMethod.version"));
        writer.print("</font></td>\r\n");
        writer.print("<td colspan=\"2\" align=\"right\" bgcolor=\"#cccccc\">");
        writer.print("<font size=\"-1\">");
        writer.print(formats[0].format(new Date()));
        writer.print("</font></td></tr>\r\n");
        writer.print("</table>\r\n");
        writer.print("</body>\r\n");
        writer.print("</html>\r\n");
        
        // Return an input stream to the underlying bytes
        writer.flush();
        
    }
    
    
    /**
     * Display an ACL list.
     *
     * @param permissionsList List of NodePermission objects
     * @param boolean Shade
     * @param writer The output will be appended to this writer
     */
    private void displayPermissions(Enumeration permissionsList,
                                    PrintWriter writer,
                                    boolean shade)
        throws IOException {
        
        if ((permissionsList != null) && (permissionsList.hasMoreElements())) {
            
            writer.print("<tr");
            if (!shade) {
                writer.print(" bgcolor=\"dddddd\"");
            } else {
                writer.print(" bgcolor=\"eeeeee\"");
            }
            writer.print(">\r\n");
            writer.print("<td align=\"left\" colspan=\"5\"><tt><b>");
            writer.print(Messages.message
                         ("org.apache.slide.webdav.GetMethod.aclinfo"));
            writer.print("</b></tt></td>\r\n");
            writer.print("</tr>\r\n");
            
            writer.print("<tr");
            if (!shade) {
                writer.print(" bgcolor=\"dddddd\"");
            } else {
                writer.print(" bgcolor=\"eeeeee\"");
            }
            writer.print(">\r\n");
            
            writer.print("<td align=\"left\" colspan=\"2\"><tt><b>");
            writer.print(Messages.message
                         ("org.apache.slide.webdav.GetMethod.subject"));
            writer.print("</b></tt></td>\r\n");
            
            writer.print("<td align=\"left\"><tt><b>");
            writer.print(Messages.message
                         ("org.apache.slide.webdav.GetMethod.action"));
            writer.print("</b></tt></td>\r\n");
            
            writer.print("<td align=\"right\"><tt><b>");
            writer.print(Messages.message
                         ("org.apache.slide.webdav.GetMethod.inheritable"));
            writer.print("</b></tt></td>\r\n");
            
            writer.print("<td align=\"right\"><tt><b>");
            writer.print(Messages.message
                         ("org.apache.slide.webdav.GetMethod.deny"));
            writer.print("</b></tt></td>\r\n");
            
/*
            writer.print("<td align=\"right\"><tt><b>");
            writer.print("<a href=\"");
            if (req.getContextPath() != null) {
                writer.print(req.getContextPath());
            }
            writer.print(managerServletPath + token.getName()
                         + "?command=addacl\">");
            writer.print(Messages.message
                         ("org.apache.slide.webdav.GetMethod.add"));
            writer.print("</a>");
            writer.print("</b></tt></td>\r\n");
*/
            
            writer.print("</tr>\r\n");
            
            while (permissionsList.hasMoreElements()) {
                
                writer.print("<tr");
                if (!shade) {
                    writer.print(" bgcolor=\"dddddd\"");
                } else {
                    writer.print(" bgcolor=\"eeeeee\"");
                }
                writer.print(">\r\n");
                
                NodePermission currentPermission =
                    (NodePermission) permissionsList.nextElement();
                
                writer.print("<td align=\"left\" colspan=\"2\"><tt>");
                writer.print(currentPermission.getSubjectUri());
                writer.print("</tt></td>\r\n");
                
                writer.print("<td align=\"left\"><tt>");
                writer.print(currentPermission.getActionUri());
                writer.print("</tt></td>\r\n");
                
                writer.print("<td align=\"right\"><tt>");
                writer.print(currentPermission.isInheritable());
                writer.print("</tt></td>\r\n");
                
                writer.print("<td align=\"right\"><tt>");
                writer.print(currentPermission.isNegative());
                writer.print("</tt></td>\r\n");
                
/*
                writer.print("<td align=\"right\"><tt><b>");
                writer.print("<a href=\"");
                if (req.getContextPath() != null) {
                    writer.print(req.getContextPath());
                }
                writer.print(managerServletPath + token.getName()
                             + "?command=removeacl"
                             + "&object=" + currentPermission.getObjectUri()
                             + "&subject=" + currentPermission.getSubjectUri()
                             + "&action=" + currentPermission.getActionUri()
                             + "\">");
                writer.print(Messages.message
                             ("org.apache.slide.webdav.GetMethod.remove"));
                writer.print("</a>");
                writer.print("</b></tt></td>\r\n");
*/
                
                writer.print("</tr>\r\n");
                    
            }
            
        }
        
    }
    
    
    /**
     * Display a lock list.
     *
     * @param locksList List of NodeLock objects
     * @param boolean Shade
     * @param writer The output will be appended to this writer
     */
    private void displayLocks(Enumeration locksList, PrintWriter writer,
                              boolean shade)
        throws IOException {
        
        if ((locksList != null) && (locksList.hasMoreElements())) {
            
            writer.print("<tr");
            if (!shade) {
                writer.print(" bgcolor=\"dddddd\"");
            } else {
                writer.print(" bgcolor=\"eeeeee\"");
            }
            writer.print(">\r\n");
            writer.print("<td align=\"left\" colspan=\"5\"><tt><b>");
            writer.print(Messages.message
                         ("org.apache.slide.webdav.GetMethod.locksinfo"));
            writer.print("</b></tt></td>\r\n");
            writer.print("</tr>\r\n");
            
            writer.print("<tr");
            if (!shade) {
                writer.print(" bgcolor=\"dddddd\"");
            } else {
                writer.print(" bgcolor=\"eeeeee\"");
            }
            writer.print(">\r\n");
            
            writer.print("<td align=\"left\"><tt><b>");
            writer.print(Messages.message
                         ("org.apache.slide.webdav.GetMethod.subject"));
            writer.print("</b></tt></td>\r\n");
            
            writer.print("<td align=\"left\"><tt><b>");
            writer.print(Messages.message
                         ("org.apache.slide.webdav.GetMethod.type"));
            writer.print("</b></tt></td>\r\n");
            
            writer.print("<td align=\"right\"><tt><b>");
            writer.print(Messages.message
                         ("org.apache.slide.webdav.GetMethod.expiration"));
            writer.print("</b></tt></td>\r\n");
            
            writer.print("<td align=\"right\"><tt><b>");
            writer.print(Messages.message
                         ("org.apache.slide.webdav.GetMethod.inheritable"));
            writer.print("</b></tt></td>\r\n");
            
            writer.print("<td align=\"right\"><tt><b>");
            writer.print(Messages.message
                         ("org.apache.slide.webdav.GetMethod.exclusive"));
            writer.print("</b></tt></td>\r\n");
            
            writer.print("</tr>\r\n");
            
            while (locksList.hasMoreElements()) {
                
                writer.print("<tr");
                if (!shade) {
                    writer.print(" bgcolor=\"dddddd\"");
                } else {
                    writer.print(" bgcolor=\"eeeeee\"");
                }
                writer.print(">\r\n");
                
                NodeLock currentLock = (NodeLock) locksList.nextElement();
                
                writer.print("<td align=\"left\"><tt>");
                writer.print(currentLock.getSubjectUri());
                writer.print("</tt></td>\r\n");
                
                writer.print("<td align=\"left\"><tt>");
                writer.print(currentLock.getTypeUri());
                writer.print("</tt></td>\r\n");
                
                writer.print("<td align=\"right\"><tt>");
                writer.print
                    (formats[0].format(currentLock.getExpirationDate()));
                writer.print("</tt></td>\r\n");
                
                writer.print("<td align=\"right\"><tt>");
                writer.print(currentLock.isInheritable());
                writer.print("</tt></td>\r\n");
                
                writer.print("<td align=\"right\"><tt>");
                writer.print(currentLock.isExclusive());
                writer.print("</tt></td>\r\n");
                
            }
            
        }
        
    }
    
    
    // -------------------------------------------------------- Private Methods
    
    
    /**
     * Render the specified file size (in bytes).
     *
     * @param size File size (in bytes)
     */
    private String renderSize(long size) {
        
        long leftSide = size / 1024;
        long rightSide = (size % 1024) / 103;   // Makes 1 digit
        if ((leftSide == 0) && (rightSide == 0) && (size > 0))
            rightSide = 1;
        
        return ("" + leftSide + "." + rightSide + " kb");
        
    }
    
    
    /**
     * Check if the conditions specified in the optional If headers are
     * satisfied.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are creating
     * @param resourceInfo File object
     * @return boolean true if the resource meets all the specified conditions,
     * and false if any of the conditions is not satisfied, in which case
     * request processing is stopped
     */
    private boolean checkIfHeaders(HttpServletRequest request,
                                     HttpServletResponse response,
                                     ResourceInfo resourceInfo)
        throws IOException {
        
        String eTag = getETag(resourceInfo, true);
        long fileLength = resourceInfo.length;
        long lastModified = resourceInfo.date;
        
        StringTokenizer commaTokenizer;
        
        String headerValue;
        
        // Checking If-Match
        headerValue = request.getHeader("If-Match");
        if (headerValue != null) {
            if (headerValue.indexOf("*") == -1) {
                
                commaTokenizer = new StringTokenizer(headerValue, ",");
                boolean conditionSatisfied = false;
                
                while (!conditionSatisfied && commaTokenizer.hasMoreTokens()) {
                    String currentToken = commaTokenizer.nextToken();
                    if (currentToken.trim().equals(eTag))
                        conditionSatisfied = true;
                }
                
                // If none of the given ETags match, 412 Precodition failed is
                // sent back
                if (!conditionSatisfied) {
                    response.sendError
                        (HttpServletResponse.SC_PRECONDITION_FAILED);
                    return false;
                }
                
            }
        }
        
        // Checking If-Modified-Since
        headerValue = request.getHeader("If-Modified-Since");
        if (headerValue != null) {
            
            // If an If-None-Match header has been specified, if modified since
            // is ignored.
            if (request.getHeader("If-None-Match") == null) {
                
                Date date = null;
                
                // Parsing the HTTP Date
                for (int i = 0; (date == null) && (i < formats.length); i++) {
                    try {
                        date = formats[i].parse(headerValue);
                    } catch (ParseException e) {
                        ;
                    }
                }
                
                if ((date != null)
                    && (lastModified <= (date.getTime() + 1000)) ) {
                    // The entity has not been modified since the date
                    // specified by the client. This is not an error case.
                    response.sendError
                        (HttpServletResponse.SC_NOT_MODIFIED);
                    return false;
                }
                
            }
            
        }
        
        // Checking If-None-Match
        headerValue = request.getHeader("If-None-Match");
        if (headerValue != null) {
            if (headerValue.indexOf("*") == -1) {
                
                commaTokenizer = new StringTokenizer(headerValue, ",");
                boolean conditionSatisfied = false;
                
                while (!conditionSatisfied && commaTokenizer.hasMoreTokens()) {
                    String currentToken = commaTokenizer.nextToken();
                    if (currentToken.trim().equals(eTag))
                        conditionSatisfied = true;
                }
                
                if (conditionSatisfied) {
                    
                    // For GET and HEAD, we should respond with
                    // 304 Not Modified.
                    // For every other method, 412 Precondition Failed is sent
                    // back.
                    if ( ("GET".equals(request.getMethod()))
                         || ("HEAD".equals(request.getMethod())) ) {
                        response.sendError
                            (HttpServletResponse.SC_NOT_MODIFIED);
                        return false;
                    } else {
                        response.sendError
                            (HttpServletResponse.SC_PRECONDITION_FAILED);
                        return false;
                    }
                }
                
            } else {
                if (resourceInfo.exists()) {
                    
                }
            }
        }
        
        // Checking If-Unmodified-Since
        headerValue = request.getHeader("If-Unmodified-Since");
        if (headerValue != null) {
            
            Date date = null;
            
            // Parsing the HTTP Date
            for (int i = 0; (date == null) && (i < formats.length); i++) {
                try {
                    date = formats[i].parse(headerValue);
                } catch (ParseException e) {
                    ;
                }
            }
            
            if ( (date != null) && (lastModified > date.getTime()) ) {
                // The entity has not been modified since the date
                // specified by the client. This is not an error case.
                response.sendError
                    (HttpServletResponse.SC_PRECONDITION_FAILED);
                return false;
            }
            
        }
        
        return true;
    }


    /**
     * Get the ETag value associated with a file.
     *
     * @param resourceInfo File object
     * @param strong True if we want a strong ETag, in which case a checksum
     * of the file has to be calculated
     */
    private String getETagValue(ResourceInfo resourceInfo, boolean strong) {
        // FIXME : Compute a strong ETag if requested, using an MD5 digest
        // of the file contents
        return resourceInfo.length + "-" + resourceInfo.date;
    }


    /**
     * Get the ETag associated with a file.
     *
     * @param resourceInfo File object
     * @param strong True if we want a strong ETag, in which case a checksum
     * of the file has to be calculated
     */
    private String getETag(ResourceInfo resourceInfo, boolean strong) {
        if (strong)
            return "\"" + getETagValue(resourceInfo, strong) + "\"";
        else
            return "W/\"" + getETagValue(resourceInfo, strong) + "\"";
    }


    /**
     * Copy the contents of the specified input stream to the specified
     * output stream, and ensure that both streams are closed before returning
     * (even in the face of an exception).
     *
     * @param istream The input stream to read from
     * @param ostream The output stream to write to
     *
     * @exception IOException if an input/output error occurs
     */
    private void copy(ResourceInfo resourceInfo,
                      InputStream resourceInputStream,
                      ServletOutputStream ostream)
    throws IOException {

        IOException exception = null;
            
        InputStream istream = new BufferedInputStream
            (resourceInputStream, input);
        
        // Copy the input stream to the output stream
        exception = copyRange(istream, ostream);
        
        // Clean up the input and output streams
        try {
            istream.close();
        } catch (Throwable t) {
            ;
        }
        
    try {
        ostream.flush();
    } catch (Throwable t) {
        ;
    }
    try {
        ostream.close();
    } catch (Throwable t) {
        ;
    }

    // Rethrow any exception that has occurred
    if (exception != null)
        throw exception;

    }


    /**
     * Copy the contents of the specified input stream to the specified
     * output stream, and ensure that both streams are closed before returning
     * (even in the face of an exception).
     *
     * @param resourceInfo The ResourceInfo object
     * @param ostream The output stream to write to
     * @param range Range the client wanted to retrieve
     * @exception IOException if an input/output error occurs
     */
    private void copy(ResourceInfo resourceInfo,
                      InputStream resourceInputStream,
                      ServletOutputStream ostream,
                      Range range)
    throws IOException {
        
        IOException exception = null;
        
        InputStream istream =
            new BufferedInputStream(resourceInputStream, input);
        exception = copyRange(istream, ostream, range.start, range.end);
        
    // Clean up the input and output streams
    try {
        istream.close();
    } catch (Throwable t) {
        ;
    }
    try {
        ostream.flush();
    } catch (Throwable t) {
        ;
    }
    try {
        ostream.close();
    } catch (Throwable t) {
        ;
    }

    // Rethrow any exception that has occurred
    if (exception != null)
        throw exception;
        
    }


    /**
     * Copy the contents of the specified input stream to the specified
     * output stream, and ensure that both streams are closed before returning
     * (even in the face of an exception).
     *
     * @param resourceInfo The ResourceInfo object
     * @param ostream The output stream to write to
     * @param ranges Enumeration of the ranges the client wanted to retrieve
     * @param contentType Content type of the resource
     * @exception IOException if an input/output error occurs
     */
    private void copy(ResourceInfo resourceInfo,
                      InputStream resourceInputStream,
                      ServletOutputStream ostream,
                      Enumeration ranges, String contentType)
    throws IOException {
        
        IOException exception = null;
        
        while ( (exception == null) && (ranges.hasMoreElements()) ) {
            
            InputStream istream =
                new BufferedInputStream(resourceInputStream, input);
        
            Range currentRange = (Range) ranges.nextElement();
            
            // Writing MIME header.
            ostream.println("--" + mimeSeparation);
            if (contentType != null)
                ostream.println("Content-Type: " + contentType);
            ostream.println("Content-Range: bytes " + currentRange.start
                           + "-" + currentRange.end + "/"
                           + currentRange.length);
            ostream.println();
            
            // Printing content
            exception = copyRange(istream, ostream, currentRange.start,
                                  currentRange.end);
            
            try {
                istream.close();
            } catch (Throwable t) {
                ;
            }
            
        }
        
        ostream.print("--" + mimeSeparation + "--");
        
    // Clean up the output streams
    try {
        ostream.flush();
    } catch (Throwable t) {
        ;
    }
    try {
        ostream.close();
    } catch (Throwable t) {
        ;
    }

    // Rethrow any exception that has occurred
    if (exception != null)
        throw exception;
        
    }


    /**
     * Copy the contents of the specified input stream to the specified
     * output stream, and ensure that both streams are closed before returning
     * (even in the face of an exception).
     *
     * @param istream The input stream to read from
     * @param ostream The output stream to write to
     * @return Exception which occured during processing
     */
    private IOException copyRange(InputStream istream,
                                  ServletOutputStream ostream) {
        
    // Copy the input stream to the output stream
    IOException exception = null;
    byte buffer[] = new byte[input];
    int len = buffer.length;
    while (true) {
        try {
                len = istream.read(buffer);
                if (len == -1)
                    break;
                ostream.write(buffer, 0, len);
        } catch (IOException e) {
        exception = e;
        len = -1;
                break;
        }
    }
        return exception;
        
    }


    /**
     * Copy the contents of the specified input stream to the specified
     * output stream, and ensure that both streams are closed before returning
     * (even in the face of an exception).
     *
     * @param istream The input stream to read from
     * @param ostream The output stream to write to
     * @param start Start of the range which will be copied
     * @param end End of the range which will be copied
     * @return Exception which occured during processing
     */
    private IOException copyRange(InputStream istream,
                                  ServletOutputStream ostream,
                                  long start, long end) {
        
        try {
            istream.skip(start);
        } catch (IOException e) {
            return e;
        }
        
    IOException exception = null;
        long bytesToRead = end - start + 1;
        
    byte buffer[] = new byte[input];
    int len = buffer.length;
    while ( (bytesToRead > 0) && (len >= buffer.length)) {
        try {
                len = istream.read(buffer);
                if (bytesToRead >= len) {
                    ostream.write(buffer, 0, len);
                    bytesToRead -= len;
                } else {
                    ostream.write(buffer, 0, (int) bytesToRead);
                    bytesToRead = 0;
                }
        } catch (IOException e) {
        exception = e;
        len = -1;
        }
        if (len < buffer.length)
        break;
    }
        
        return exception;
        
    }


    /**
     * Parse the range header.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are creating
     * @return Vector of ranges
     */
    private Vector parseRange(HttpServletRequest request,
                              HttpServletResponse response,
                              ResourceInfo resourceInfo)
        throws IOException {
        
        // Checking If-Range
        String headerValue = request.getHeader("If-Range");
        if (headerValue != null) {
            
            String eTag = getETag(resourceInfo, true);
            long lastModified = resourceInfo.date;
            
            Date date = null;
            
            // Parsing the HTTP Date
            for (int i = 0; (date == null) && (i < formats.length); i++) {
                try {
                    date = formats[i].parse(headerValue);
                } catch (ParseException e) {
                    ;
                }
            }
            
            if (date == null) {
                
                // If the ETag the client gave does not match the entity
                // etag, then the entire entity is returned.
                if (!eTag.equals(headerValue.trim()))
                    return null;
                
            } else {
                
                // If the timestamp of the entity the client got is older than
                // the last modification date of the entity, the entire entity
                // is returned.
                if (lastModified > (date.getTime() + 1000))
                    return null;
                
            }
            
        }
        
        long fileLength = resourceInfo.length;
        
        if (fileLength == 0)
            return null;
        
        // Retrieving the range header (if any is specified
        String rangeHeader = request.getHeader("Range");
        
        if (rangeHeader == null)
            return null;
        // bytes is the only range unit supported (and I don't see the point
        // of adding new ones).
        if (!rangeHeader.startsWith("bytes")) {
            response.sendError
                (HttpServletResponse.SC_REQUESTED_RANGE_NOT_SATISFIABLE);
            return null;
        }
        
        rangeHeader = rangeHeader.substring(6);
        
        // Vector which will contain all the ranges which are successfully
        // parsed.
        Vector result = new Vector();
        StringTokenizer commaTokenizer = new StringTokenizer(rangeHeader, ",");
        
        // Parsing the range list
        while (commaTokenizer.hasMoreTokens()) {
            String rangeDefinition = commaTokenizer.nextToken();
            
            Range currentRange = new Range();
            currentRange.length = fileLength;
            
            int dashPos = rangeDefinition.indexOf('-');
            
            if (dashPos == -1) {
                response.sendError
                    (HttpServletResponse.SC_REQUESTED_RANGE_NOT_SATISFIABLE);
                return null;
            }
            
            if (dashPos == 0) {
                
                try {
                    long offset = Long.parseLong(rangeDefinition);
                    currentRange.start = fileLength + offset;
                    currentRange.end = fileLength - 1;
                } catch (NumberFormatException e) {
                    response.sendError
                        (HttpServletResponse
                         .SC_REQUESTED_RANGE_NOT_SATISFIABLE);
                    return null;
                }
                
            } else {
                
                try {
                    currentRange.start = Long.parseLong
                        (rangeDefinition.substring(0, dashPos));
                    if (dashPos < rangeDefinition.length() - 1)
                        currentRange.end = Long.parseLong
                            (rangeDefinition.substring
                             (dashPos + 1, rangeDefinition.length()));
                    else
                        currentRange.end = fileLength - 1;
                } catch (NumberFormatException e) {
                    response.sendError
                        (HttpServletResponse
                         .SC_REQUESTED_RANGE_NOT_SATISFIABLE);
                    return null;
                }
                
            }
            
            if (!currentRange.validate()) {
                response.sendError
                    (HttpServletResponse.SC_REQUESTED_RANGE_NOT_SATISFIABLE);
                return null;
            }
            
            result.addElement(currentRange);
        }
        
        return result;
    }


    // ------------------------------------------------------ Range Inner Class
    
    
    private class Range {
        
        public long start;
        public long end;
        public long length;
        
        /**
         * Validate range.
         */
        public boolean validate() {
            return ( (start >= 0) && (end >= 0) && (length > 0)
                     && (start <= end) && (end < length) );
        }
        
    }
    
    
    // ----------------------------------------------  ResourceInfo Inner Class
    
    
    private class ResourceInfo {
        
        
        /**
         * Constructor.
         *
         * @param pathname Path name of the file
         */
        public ResourceInfo(String path, NodeRevisionDescriptor properties) {
            
            this.path = path;
            this.exists = true;
            this.creationDate = properties.getCreationDateAsDate().getTime();
            this.date = properties.getLastModifiedAsDate().getTime();
            this.httpDate = properties.getLastModified();
            this.length = properties.getContentLength();
            
        }
        
        
        public String path;
        public long creationDate;
        public String httpDate;
        public long date;
        public long length;
        //public boolean collection;
        public boolean exists;
        
        
        /**
         * Test if the associated resource exists.
         */
        public boolean exists() {
            return exists;
        }
        
        
        /**
         * String representation.
         */
        public String toString() {
            return path;
        }
        
        
    }
    
    
}
